/*
   acpi.c - access to ACPI BIOS structures
   
   FanIO driver V2.0
   Copyright(c) 2001-2006 Christian Diefer

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

static PUCHAR rsdtVA=NULL;
static PUCHAR xsdtVA=NULL;
static PUCHAR rsdtPA, xsdtPA;

static int findRSDP( PUCHAR addr, ULONG size, char* pIOBuffer );
static void analyzeRSDT( char* pIOBuffer );
static void analyzeXSDT( char* pIOBuffer );
static void remapACPITable( PUCHAR virt, PUCHAR phys, PUCHAR* dest, char* pIOBuffer );

/////////////////////////////////////////////////////////////////////////////
int acpiInit( char* pIOBuffer )
{
  PHYSICAL_ADDRESS lowPA;
  PUCHAR lowVA;
  static int found=0;
  static int initialized=0;     // ensure acpiInit() is called only once

  *pIOBuffer = '\0';

//  if( initialized )
//    return found;

  cleanUp();

  initMappings( biosVA );

  // search two memory locations according to ACPI spec:
  initialized = -1;
  lowPA.LowPart = 0;
  lowPA.HighPart = 0;

#ifdef ACPI_DEBUG
  strcat( pIOBuffer, "Searching EBDA for signature...\r\n" );
#endif

  // 1. search first 1024 bytes of physical memory
  lowVA = MmMapIoSpace( lowPA, EBDA_SIZE, MmNonCached );
  found = findRSDP( lowVA, EBDA_SIZE, pIOBuffer );
  MmUnmapIoSpace( lowVA, EBDA_SIZE );

  if( !found )
  {
    // 2. search whole BIOS area (0xe0000-0xfffff)
#ifdef ACPI_DEBUG
    strcat( pIOBuffer, "Searching BIOS area for signature...\r\n" );
#endif

    found = findRSDP( biosVA, BIOSAREA_SIZE, pIOBuffer );
  }

  if( !found )
  {
    strcat( pIOBuffer, "*** ERROR: No ACPI BIOS found ***\r\n" );
    return 0;      
  }

  if( rsdtVA )
    analyzeRSDT( pIOBuffer );
  else if( xsdtVA )
    analyzeXSDT( pIOBuffer );

  dumpMappings( pIOBuffer );
  
  return -1;
}

/////////////////////////////////////////////////////////////////////////////
void acpiExit()
{
  cleanUp();
}

/////////////////////////////////////////////////////////////////////////////
static int findRSDP( PUCHAR addr, ULONG size, char* pIOBuffer )
{
  PUCHAR cur;

#ifdef ACPI_DEBUG
  sprintf( &pIOBuffer[strlen(pIOBuffer)], "Searching VA 0x%08p, %d Bytes...\r\n", addr, size );
#endif

  for( cur=addr; cur<addr+size; cur+=SIG_STEP )
  {
    if( (*(PULONG)cur)==RSDP_SIG1 && (*(PULONG)(cur+4))==RSDP_SIG2 
    &&  (cur[15]==0 || cur[15]==2) )      // check for sig, ACPI V1+V2 only
    {
      UCHAR x = 0;
      int j;

#ifdef ACPI_DEBUG
      sprintf( &pIOBuffer[strlen(pIOBuffer)], "Sig found at VA 0x%08p\r\n", cur );
#endif

      for( j=0; j<20; j++ )
        x += cur[j];

      if( x == 0 )
      {
        if( cur[15] == 0 )     // ACPI V1
        {
#ifdef _WIN64
          rsdtPA = (PUCHAR)(ULONG64)*((PULONG)(cur+16));
          rsdtVA = mapPhysicalToVirtual( rsdtPA, SDT_HDRSIZE );
#else
          rsdtPA = (PUCHAR)*((PULONG)(cur+16));
          rsdtVA = mapPhysicalToVirtual( rsdtPA, SDT_HDRSIZE );
#endif

#ifdef ACPI_DEBUG
          sprintf( pIOBuffer+strlen(pIOBuffer), "ACPI V1 BIOS detected\r\nRSDT @ VA 0x%08p (PA 0x%08p)\r\n", 
                   rsdtVA, rsdtPA );
#endif
        }
        else                    // ACPI V2
        {
          if( *((PULONG)(cur+28)) != 0 )
          {
            strcat( pIOBuffer, "*** ERROR: XSDT highpart in RSDP is not zero ! ***" );
            return 0;
          }

#ifdef _WIN64
          xsdtPA = (PUCHAR)(ULONG64)*((PULONG)(cur+24));
          xsdtVA = mapPhysicalToVirtual( xsdtPA, SDT_HDRSIZE );
#else
          xsdtPA = (PUCHAR)*((PULONG)(cur+24));
          xsdtVA = mapPhysicalToVirtual( xsdtPA, SDT_HDRSIZE );
#endif

#ifdef ACPI_DEBUG
          sprintf( pIOBuffer+strlen(pIOBuffer), "ACPI V2 BIOS detected\r\nXSDT @ VA 0x%08p (PA 0x%08p)\r\n", 
                   xsdtVA, xsdtPA );
#endif
        }

        return -1;      // signature found, checksum ok
      }
    }    
  }

  return 0;
}

/////////////////////////////////////////////////////////////////////////////
static void analyzeRSDT( char* pIOBuffer )
{
  ULONG size = *(PULONG)(rsdtVA+4);
  ULONG offset = 0;
  
#ifdef ACPI_DEBUG
  strcat( pIOBuffer, "Analyzing RSDT...\r\n" );
#endif

  remapACPITable( rsdtVA, rsdtPA, &rsdtVA, pIOBuffer );
  
  size -= SDT_HDRSIZE;

  if( size % 4 )
  {
#ifdef ACPI_DEBUG
    strcat( pIOBuffer, "RSDT size not DWORD aligned\r\n" );
#endif
    return;
  }

  while( offset < size )
  {
#ifdef _WIN64
    PUCHAR phys = (PUCHAR)(ULONG64)*(PULONG)(rsdtVA+36+offset);
#else
    PUCHAR phys = (PUCHAR)*(PULONG)(rsdtVA+36+offset);
#endif

    PUCHAR virt = mapPhysicalToVirtual( phys, SDT_HDRSIZE );

#ifdef ACPI_DEBUG
    sprintf( pIOBuffer+strlen(pIOBuffer), "Table found at PA 0x%08p...\r\n", phys );
#endif

    remapACPITable( virt, phys, &virt, pIOBuffer );

    // process new table
    // ...

    offset += 4;
  }

}

/////////////////////////////////////////////////////////////////////////////
static void analyzeXSDT( char* pIOBuffer )
{
  ULONG size = *(PULONG)(xsdtVA+4);
  ULONG offset = 0;

#ifdef ACPI_DEBUG
  strcat( pIOBuffer, "Analyzing XSDT...\r\n" );
#endif

  remapACPITable( xsdtVA, xsdtPA, &xsdtVA, pIOBuffer );

  size -= SDT_HDRSIZE;
  
  if( size % 8 )
  {
#ifdef ACPI_DEBUG
    strcat( pIOBuffer, "XSDT size not DWORD aligned\r\n" );
#endif
    return;
  }

  while( offset < size )
  {
#ifdef _WIN64
    PUCHAR phys = (PUCHAR)(ULONG64)*(PULONG)(xsdtVA+36+offset);
#else
    PUCHAR phys = (PUCHAR)*(PULONG)(xsdtVA+36+offset);
#endif

    PUCHAR virt = mapPhysicalToVirtual( phys, SDT_HDRSIZE );

#ifdef ACPI_DEBUG
    sprintf( pIOBuffer+strlen(pIOBuffer), "Table found at PA 0x%08p ...\r\n", phys );
#endif

    remapACPITable( virt, phys, &virt, pIOBuffer );
    
    // process new table
    // ...

    offset += 8;
  }
  

}

/////////////////////////////////////////////////////////////////////////////
static void remapACPITable( PUCHAR virt, PUCHAR phys, PUCHAR* dest, char* pIOBuffer )
{
  ULONG size = *(PULONG)(virt+4);

#ifdef ACPI_DEBUG
  char sig[12];

  strncpy( sig, virt, 4 );
  sig[4] = '\0';
  sprintf( pIOBuffer+strlen(pIOBuffer), "  Remapping '%s', %u Bytes from VA 0x%08p\r\n", sig, size, virt );

  strncpy( sig, virt+10, 6 );
  sig[6] = '\0';
  sprintf( pIOBuffer+strlen(pIOBuffer), "  OEMID:'%s'\r\n", sig );

  strncpy( sig, virt+16, 8 );
  sig[8] = '\0';
  sprintf( pIOBuffer+strlen(pIOBuffer), "  OEM Table ID:'%s'\r\n", sig );
#endif

  *dest = changeMapping( phys, SDT_HDRSIZE, size );
  
#ifdef ACPI_DEBUG
  sprintf( pIOBuffer+strlen(pIOBuffer), "  Table remapped to VA 0x%08p (PA 0x%08p)\r\n", *dest, phys );
#endif
}

/////////////////////////////////////////////////////////////////////////////
#include "acpimem.c"
