/*
   acpimem.c - handle physical -> virtual address space mappings
   
   FanIO driver V2.0
   Copyright(c) 2001-2006 Christian Diefer

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/////////////////////////////////////////////////////////////////////////////
void initMappings( PUCHAR vir )
{
  // the BIOS area is always mapped
  pva_mappings[0].phys = (PUCHAR)0xe0000;
  pva_mappings[0].virt = vir;
  pva_mappings[0].size = 0x20000;
  pva_mappings[0].unmap = 0;    // do not unmap BIOS area automatically
  nextEntry = 1;
}

/////////////////////////////////////////////////////////////////////////////
PUCHAR mapPhysicalToVirtual( PUCHAR phys, ULONG size )
{
  PUCHAR v; 
  PHYSICAL_ADDRESS x;
  int found=0;
  int i;

  for( i=0; i<nextEntry; i++ )
  {
    if( phys < pva_mappings[i].phys )
      continue;

    if( phys+size > pva_mappings[i].phys+pva_mappings[i].size )
      continue;

    found = -1;         // block fits in current mapping
    break;
  }

  if( !found )
  {
#ifdef _WIN64
    x.HighPart = (ULONG)(((ULONG64)phys)>>32);
    x.LowPart = (ULONG)(((ULONG64)phys)&0xffffffff);
#else
    x.HighPart = 0;
    x.LowPart = (ULONG)phys;
#endif
  
    v = MmMapIoSpace( x, size, MmNonCached );
    
    pva_mappings[nextEntry].phys = phys;
    pva_mappings[nextEntry].virt = v;
    pva_mappings[nextEntry].size = size;
    pva_mappings[nextEntry].unmap = -1;
    nextEntry++;
  }
  else
  {
    v = (PUCHAR)(phys - pva_mappings[i].phys);
    v += (ULONG64)pva_mappings[i].virt;  
  }

  return (PUCHAR)v;
}

/////////////////////////////////////////////////////////////////////////////
PUCHAR changeMapping( PUCHAR phys, ULONG oldsize, ULONG newsize )
{
  int i;
  int found=0;
  PUCHAR base;
  PHYSICAL_ADDRESS ad;
  
  
  for( i=0; i<nextEntry; i++ )
  {
    if( phys==pva_mappings[i].phys && oldsize==pva_mappings[i].size )
    {
      found = -1;
      break;
    }
  }

  if( !found )
    return mapPhysicalToVirtual( phys, newsize );

  MmUnmapIoSpace( pva_mappings[i].virt, oldsize );

#ifdef _WIN64
  ad.HighPart = (ULONG)(((ULONG64)phys)>>32);
  ad.LowPart = (ULONG)(((ULONG64)phys)&0xffffffff);
#else
  ad.HighPart = 0;
  ad.LowPart = (ULONG)phys;
#endif

  base = MmMapIoSpace( ad, newsize, MmNonCached );

  pva_mappings[i].phys = phys;
  pva_mappings[i].virt = base;
  pva_mappings[i].size = newsize;
  pva_mappings[i].unmap = -1;  

  return base;
}

/////////////////////////////////////////////////////////////////////////////
void cleanUp()
{
  int i;
  
  for( i=0; i<nextEntry; i++ )
    if( pva_mappings[i].unmap )
      MmUnmapIoSpace( pva_mappings[i].virt, pva_mappings[i].size );  
  
  nextEntry = 0;      
}

/////////////////////////////////////////////////////////////////////////////
void dumpMappings( char* pIOBuffer )
{
  int i;
  
  strcat( pIOBuffer, "\r\nPA->VA Mappings:\r\n" );

  for( i=0; i<nextEntry; i++ )
  {
    sprintf( pIOBuffer+strlen(pIOBuffer), "PA 0x%08p -> VA 0x%08p (%d Bytes)\r\n", 
             pva_mappings[i].phys, pva_mappings[i].virt, pva_mappings[i].size );
  }
}

/////////////////////////////////////////////////////////////////////////////
