/*
   diagio.c - Dell Inspiron 8x00 access functions for SMBIOS diagnostics interface

   FanIO driver V2.x
   Copyright(c) 2001-2007 Christian Diefer

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

ULONG getFanSpeedIO( ULONG fanNo, PULONG pIOBuffer )
{
  int x;
  SMBIOSCMD sc;
  sc.cmd = 0x00a3;
  sc.data = fanNo;
  sc.stat1 = 0;
  sc.stat2 = 0;

  x = smfunction( &sc );
  
  pIOBuffer[1] = sc.cmd;
  pIOBuffer[2] = sc.data;
  pIOBuffer[3] = sc.stat1;
  pIOBuffer[4] = sc.stat2;
  
  if( x != 1 )
    return 0xffffffff;

  return sc.cmd; 
}

ULONG setFanSpeedIO( ULONG fanNo, ULONG speed, PULONG pIOBuffer )
{
  int x;
  SMBIOSCMD sc;
  sc.cmd = 0x1a3;
  sc.data = (speed<<8) | fanNo;
  sc.stat1 = 0;
  sc.stat2 = 0;

  x = smfunction( &sc );
  
  pIOBuffer[1] = sc.cmd;
  pIOBuffer[2] = sc.data;
  pIOBuffer[3] = sc.stat1;
  pIOBuffer[4] = sc.stat2;
  
  if( x != 1 )
    return 0xffffffff;

  return 0;
}

ULONG getFanRPMIO( ULONG fanNo, PULONG pIOBuffer )
{
  int x;
  SMBIOSCMD sc;
  sc.cmd = 0x02a3;
  sc.data = fanNo;
  sc.stat1 = 0;
  sc.stat2 = 0;

  x = smfunction( &sc );
  
  pIOBuffer[1] = sc.cmd;
  pIOBuffer[2] = sc.data;
  pIOBuffer[3] = sc.stat1;
  pIOBuffer[4] = sc.stat2;
  
  if( x != 1 )
    return 0xffffffff;

  return sc.cmd * 30;
}

ULONG getTempIO( ULONG thermNo, PULONG pIOBuffer )
{
  int x;
  SMBIOSCMD sc;
  sc.cmd = 0x10a3;
  sc.data = thermNo;
  sc.stat1 = 0;
  sc.stat2 = 0;

  x = smfunction( &sc );

  pIOBuffer[1] = sc.cmd;
  pIOBuffer[2] = sc.data;
  pIOBuffer[3] = sc.stat1;
  pIOBuffer[4] = sc.stat2;
  
  if( x != 1 )
    return 0xffffffff;

  return sc.cmd;  
}

ULONG execSMMOp( ULONG cmd, ULONG data, ULONG stat1, ULONG stat2, PULONG pIOBuffer )
{
  int x;
  SMBIOSCMD sc;
  sc.cmd = cmd;
  sc.data = data;
  sc.stat1 = stat1;
  sc.stat2 = stat2;

  x = smfunction( &sc );

  pIOBuffer[1] = sc.cmd;
  pIOBuffer[2] = sc.data;
  pIOBuffer[3] = sc.stat1;
  pIOBuffer[4] = sc.stat2;
  
  if( x != 1 )
    return 0xffffffff;

  return sc.cmd;  
}

/*
  -----------------------------------------
  smfunction: call SMBIOS function

  In: control structure

  Out: 1:ok 0:error

  According to the module NBSVC.MDM the following SMM (System Management Mode) functions are currently available (V1266):

  cmd     function
  0x00a3  get current speed indicator of a fan (args: fan)
  0x01a3  set speed of a fan (args: fan & speed)
  0x02a3  get RPM of a fan (args: fan)
  0x03a3  ??? (1 byte)
  0x04a3  get nominal fan speed (2 args)
  0x05a3  get fan tolerance speed (2 args)
  0x10a3  get sensor temperature (1 arg: sensor#)
  0x11a3  ???
  0x12a3  arg 0x0003=NBSVC-Query 
          arg 0x0000=NBSVC-Clear
          arg 0x122=NBSVC-Start Trend
          arg 0x0100=NBSVC-Stop Trend
          arg 0x02??=NBSVC-Read
  0x21a3  ??? (2 args: 1 byte (oder 0x16) + 1 byte)
  0x22a3  get charger info (1 arg)
  0x23a3  ??? (4 args: 2x 1 byte, 1xword, 1xdword)
  0x24a3  get adaptor info status (1 arg oder 0x03)
  0x30a3  ??? (no args)
  0x31a3  ??? (no args)
  0x32a3  ??? (no args)
  0x33a3  ??? (no args)
  0x36a3  get hotkey scancode list (args see diags)
  0x37a3  ??? (no args)
  0x40a3  get docking state (no args)
  0xf0a3  ??? (2 args)
  0xfea3  check SMBIOS version (1 arg)
  0xffa3  check SMBIOS interface (returns:"DELLDIAG")
  -------------------------------------------
*/
#ifndef _WIN64
int smfunction( SMBIOSCMD* sc )
{
  int status = 1;
  ULONG cmd = sc->cmd;
  ULONG data = sc->data;
  ULONG stat1 = sc->stat1;
  ULONG stat2 = sc->stat2;

  _asm
  {
    pushad
    mov eax,cmd
    mov ebx,data
    mov ecx,stat1
    mov edx,stat2
    out 0b2h,al
    out 084h,al
    mov cmd,eax
    mov data,ebx
    mov stat1,ecx
    mov stat2,edx
    popad
  }

  // did it work ?
  if( (sc->cmd==cmd && sc->data==data && sc->stat1==stat1 && sc->stat2==stat2)  // function not supported
  ||  ((cmd&0xffff)==0xffff) )     // or invalid device
    status = 0;

  sc->cmd = cmd;
  sc->data = data;
  sc->stat1 = stat1;
  sc->stat2 = stat2;

  return status;
}
#endif
