/*
   fanio.c - Dell Inspiron/Latitude/Precision fan io driver V2 for Windows NT4/2000/XP

   FanIO driver V2.x
   Copyright(c) 2001-2007 Christian Diefer
*/

#include "fanio.h"

//--------------------------------------------------
// driver entry
//--------------------------------------------------
NTSTATUS DriverEntry( IN PDRIVER_OBJECT DriverObject, IN PUNICODE_STRING RegistryPath )
{
  PDEVICE_OBJECT DeviceObject;
  NTSTATUS mStatus;
  UNICODE_STRING uniName, uniDOSName;
  
  // get access to whole BIOS area
  biosPA.LowPart = 0xe0000;
  biosPA.HighPart = 0;
  biosVA = MmMapIoSpace( biosPA, 0x20000, MmNonCached );
  
  vbiosVA = NULL;

  // register device functions
  DriverObject->MajorFunction[IRP_MJ_CREATE] = FanIO_DispatchCreateClose;
  DriverObject->MajorFunction[IRP_MJ_CLOSE] = FanIO_DispatchCreateClose;
  DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = FanIO_DispatchDeviceControl;
  DriverObject->DriverUnload = FanIO_Unload;

  // translate device strings to unicode
  RtlInitUnicodeString( &uniName, FANIO_DEVICE_NAME );
  RtlInitUnicodeString( &uniDOSName, DOS_DEVICE_NAME );

  // create driver interface
  mStatus = IoCreateDevice( DriverObject,
                            0,
                            &uniName,
                            FILE_DEVICE_UNKNOWN,
                            0,
                            FALSE,      // share driver object among Win32 processes
                            &DeviceObject );

  if( !NT_SUCCESS(mStatus) )
  {
    DbgPrint(("FanIO: Can't create device object !\n"));
    return mStatus;
  }

  // create symlink
  mStatus = IoCreateSymbolicLink( &uniDOSName, &uniName );
  if( !NT_SUCCESS(mStatus) )
  {
    DbgPrint(("FanIO: Can't create link\n"));
    return mStatus;
  }

  // we are doing only buffered io's
  DeviceObject->Flags |= DO_BUFFERED_IO;

  DbgPrint(("FanIO: DriverEntry successful\n"));

  return STATUS_SUCCESS;
}

//--------------------------------------------------
// handle CreateFile & CloseHandle (always return success)
//--------------------------------------------------
NTSTATUS FanIO_DispatchCreateClose( IN PDEVICE_OBJECT devObj, IN PIRP Irp )
{
  DbgPrint(("FanIO: Open / Close\n"));

  Irp->IoStatus.Information = 0;
  Irp->IoStatus.Status = STATUS_SUCCESS;
  IoCompleteRequest( Irp, IO_NO_INCREMENT );

  return STATUS_SUCCESS;
}

//--------------------------------------------------
// the dispatch function for ioctl's
//--------------------------------------------------
NTSTATUS FanIO_DispatchDeviceControl( IN PDEVICE_OBJECT devObj, IN PIRP Irp )
{
  PULONG pIOBuffer;             // buffered memory for communication with process
  ULONG InBufferSize;           // actual input buffer size
  ULONG OutBufferSize;          // actual output buffer size
  ULONG ExpInBufferSize;        // expected input buffer size
  ULONG ExpOutBufferSize;       // expected input buffer size
  ULONG IoctlCode;              // Ioctl code from process
  USHORT fanNo, speed;          // fan no and speed for input from buffer
  USHORT thermNo, speedIndex;   // thermistor# for input from buffer
  ULONG data;                   // return value (speed, temperature, etc.) for output
  SMBIOSCMD sc;                 // SMBIOS communication structure
  PIO_STACK_LOCATION IrpStack = IoGetCurrentIrpStackLocation( Irp );

  // retrieve IO control code
  IoctlCode = IrpStack->Parameters.DeviceIoControl.IoControlCode;

  // retrieve actual buffer lengths
  InBufferSize = IrpStack->Parameters.DeviceIoControl.InputBufferLength;
  OutBufferSize = IrpStack->Parameters.DeviceIoControl.OutputBufferLength;

  // NT copies inbuf here before entry and copies this to outbuf after return,
  // for METHOD_BUFFERED IOCTL's (one buffer for input and output)
  pIOBuffer = (PULONG)Irp->AssociatedIrp.SystemBuffer;

  Irp->IoStatus.Information = 0;    // default return value for IoStatus

  switch( IoctlCode )
  {
    case IOCTL_FANIO_SET_FAN_SPEED:
        ExpInBufferSize = sizeof(ULONG);
        ExpOutBufferSize = 5 * sizeof(ULONG);
        break;

    case IOCTL_FANIO_GET_FAN_SPEED:
        ExpInBufferSize = sizeof(ULONG);
        ExpOutBufferSize = 5 * sizeof(ULONG);
        break;

    case IOCTL_FANIO_GET_FAN_RPM:
        ExpInBufferSize = sizeof(ULONG);
        ExpOutBufferSize = 5 * sizeof(ULONG);
        break;

    case IOCTL_FANIO_GET_TEMP:
        ExpInBufferSize = sizeof(ULONG);
        ExpOutBufferSize = 5 * sizeof(ULONG);
        break;

    case IOCTL_FANIO_GET_VERSION:
    case IOCTL_FANIO_GET_VERSION_LEGACY:
        ExpInBufferSize = 0;
        ExpOutBufferSize = 2 * sizeof(ULONG);
        break;

    case IOCTL_FANIO_GET_SYSTEM_INFO:
        ExpInBufferSize = 0;
        ExpOutBufferSize = 16384;    // 16K temporary buffer for system info
        break;

    case IOCTL_FANIO_GET_CPU_SPEED:
        ExpInBufferSize = 0;
        ExpOutBufferSize = 8;    // 8 bytes for speed string
        break;

    case IOCTL_FANIO_CHECK_SENSOR:
        ExpInBufferSize = 0;
        ExpOutBufferSize = SENSOR_BUFSIZE * sizeof(ULONG);
        break;

    case IOCTL_FANIO_SET_SENSOR_SPEED:
        ExpInBufferSize = sizeof(ULONG);
        ExpOutBufferSize = sizeof(ULONG);
        break;

    case IOCTL_FANIO_GET_SENSOR_SPEED:
        ExpInBufferSize = sizeof(ULONG);
        ExpOutBufferSize = sizeof(ULONG);
        break;

    case IOCTL_FANIO_ACPI_INFO:
        ExpInBufferSize = 0;
        ExpOutBufferSize = 32768;    // 32K temporary buffer for ACPI info
        break;
        
    case IOCTL_FANIO_GET_SENSOR_MAX_FANS:
        ExpInBufferSize = 0;
        ExpOutBufferSize = sizeof(ULONG);
        break;

   case IOCTL_FANIO_MAP_VBIOS:
        ExpInBufferSize = 0;
        ExpOutBufferSize = 4*sizeof(ULONG);   // [0]: 0=okay, <>0=Fehler [1]:VA des gemappten VBIOS(Lowpart), [2]=VA Highpart, [3]=Gre des VBIOS in 512Byte-Seiten
        break;

   case IOCTL_FANIO_UNMAP_VBIOS:
        ExpInBufferSize = 0;
        ExpOutBufferSize = 0;
        break;

   case IOCTL_FANIO_GET_VBIOS_CHUNK:
        ExpInBufferSize = sizeof(ULONG);    // Nr. der VBIOS-Seite (0 ... vbiosSize-1)
        ExpOutBufferSize = 512;             // 512=Gre einer Seite
        break;
        
   case IOCTL_FANIO_READ_MSR:
        ExpInBufferSize = sizeof(ULONG);    // msr
        ExpOutBufferSize = 2*sizeof(ULONG); // 8 Byte (UINT64) als Rckgabe
        break;

    default:
        return STATUS_INVALID_PARAMETER;
  }

  // check buffer sizes
  if( InBufferSize<ExpInBufferSize || OutBufferSize<ExpOutBufferSize )
    return STATUS_BUFFER_TOO_SMALL;

  fanNo = 0;
  speed = 0;
  thermNo = 0;
  speedIndex = 0;

  // get & check input parameters:
  switch( IoctlCode )           // ioctl dependend
  {
    case IOCTL_FANIO_SET_FAN_SPEED:
        fanNo = ((PUSHORT)pIOBuffer)[0];
        speed = ((PUSHORT)pIOBuffer)[1];
        break;

    case IOCTL_FANIO_GET_FAN_SPEED:
        fanNo = (USHORT)(pIOBuffer[0]);
        break;

    case IOCTL_FANIO_GET_FAN_RPM:
        fanNo = (USHORT)(pIOBuffer[0]);
        break;

    case IOCTL_FANIO_GET_TEMP:
        thermNo = ((PUSHORT)pIOBuffer)[0];
        break;

    case IOCTL_FANIO_GET_VERSION:
    case IOCTL_FANIO_GET_VERSION_LEGACY:
    case IOCTL_FANIO_GET_SYSTEM_INFO:
    case IOCTL_FANIO_GET_CPU_SPEED:
    case IOCTL_FANIO_CHECK_SENSOR:
    case IOCTL_FANIO_ACPI_INFO:
    case IOCTL_FANIO_GET_SENSOR_MAX_FANS:
    case IOCTL_FANIO_MAP_VBIOS:
    case IOCTL_FANIO_UNMAP_VBIOS:
    case IOCTL_FANIO_GET_VBIOS_CHUNK:
    case IOCTL_FANIO_READ_MSR:
        break;

    case IOCTL_FANIO_SET_SENSOR_SPEED:
        fanNo = ((PUSHORT)pIOBuffer)[0];
        speedIndex = ((PUSHORT)pIOBuffer)[1];
        break;

    case IOCTL_FANIO_GET_SENSOR_SPEED:
        fanNo = (USHORT)(pIOBuffer[0]);
        break;

    default:
        return STATUS_INVALID_PARAMETER;
  }

  if( fanNo>3 || speed>2 || thermNo>255 || speedIndex>63 )
    return STATUS_INVALID_PARAMETER;

  // execute the respective function as indicated by ioctl
  switch( IoctlCode )
  {
    case IOCTL_FANIO_GET_FAN_SPEED:
        *pIOBuffer = getFanSpeedIO( fanNo, pIOBuffer );
        break;

    case IOCTL_FANIO_SET_FAN_SPEED:
        *pIOBuffer = setFanSpeedIO( fanNo, speed, pIOBuffer );
        break;

    case IOCTL_FANIO_GET_FAN_RPM:
        *pIOBuffer = getFanRPMIO( fanNo, pIOBuffer );
        break;

    case IOCTL_FANIO_GET_TEMP:
        *pIOBuffer = getTempIO( thermNo, pIOBuffer );
        break;

    case IOCTL_FANIO_GET_VERSION:
    case IOCTL_FANIO_GET_VERSION_LEGACY:
        pIOBuffer[0] = MAJORVERSION;
        pIOBuffer[1] = MINORVERSION;
        break;

    case IOCTL_FANIO_GET_SYSTEM_INFO:
        getSMBIOS( (char*)pIOBuffer );
        break;

    case IOCTL_FANIO_GET_CPU_SPEED:
        getCPUSpeed( (char*)pIOBuffer );
        break;

    case IOCTL_FANIO_CHECK_SENSOR:
        *pIOBuffer = detectFanSensorLL( pIOBuffer );
        break;

    case IOCTL_FANIO_SET_SENSOR_SPEED:
        *pIOBuffer = setFanSpeedLL( fanNo, speedIndex );
        break;

    case IOCTL_FANIO_GET_SENSOR_SPEED:
        *pIOBuffer = getFanSpeedLL( fanNo );
        break;

    case IOCTL_FANIO_ACPI_INFO:
        acpiInit( (char*)pIOBuffer );
        break;
        
    case IOCTL_FANIO_GET_SENSOR_MAX_FANS:
        *pIOBuffer = getMaxFanNoLL();
        break;

   case IOCTL_FANIO_MAP_VBIOS:
        mapVBIOS( pIOBuffer );
        break;

   case IOCTL_FANIO_UNMAP_VBIOS:
        unmapVBIOS();
        break;

   case IOCTL_FANIO_GET_VBIOS_CHUNK:
        getVBIOSChunk( pIOBuffer );
        break;

    case IOCTL_FANIO_READ_MSR:
#ifdef _WIN64
        {
          UINT64 x = readMSR( *pIOBuffer );
          memcpy( pIOBuffer, &x, 8 );
        }
#else
        __asm {
          push ebx
          mov ebx,pIOBuffer
          mov ecx,dword ptr [ebx]
          rdmsr
          mov dword ptr [ebx],edx
          mov dword ptr [ebx+4],eax
          pop ebx
        }
#endif
        break;
        
    default:
        return STATUS_INVALID_PARAMETER;
  }

  // return # of bytes written
  Irp->IoStatus.Information = ExpOutBufferSize;

  IoCompleteRequest( Irp, IO_NO_INCREMENT );

  return STATUS_SUCCESS;
}

//--------------------------------------------------
// remove driver and release resources (links)
//--------------------------------------------------
VOID FanIO_Unload( IN PDRIVER_OBJECT DriverObject )
{
  UNICODE_STRING uniLinkName;
  NTSTATUS mStatus;

  DbgPrint(("FanIO: Entering unload function\n"));

  acpiExit();

  // terminate all PA->VA mappings
  if( biosVA != NULL )
    MmUnmapIoSpace( biosVA, 0x20000 );

  if( biosVA2Mapped )
    MmUnmapIoSpace( biosVA2, smTableLength );
    
  if( vbiosVA != NULL )
    MmUnmapIoSpace( vbiosVA, vbiosSize*512 );   

  RtlInitUnicodeString( &uniLinkName, DOS_DEVICE_NAME );

  mStatus = IoDeleteSymbolicLink( &uniLinkName );
  if( !NT_SUCCESS(mStatus) )
  {
    DbgPrint(("FanIO: Can't remove link\n"));
  }
  else
  {
    DbgPrint(("FanIO: Link removed\n"));
    IoDeleteDevice( DriverObject->DeviceObject );
  }

  DbgPrint(("FanIO: End of unload\n"));
}

//--------------------------------------------------
#include "smbios.c"
#include "lpc.c"
#include "diagio.c"
#include "acpi.c"
#include "vbios.c"
