/*
   smbios.c - access to SMBIOS table
   
   FanIO driver V2.x
   Copyright(c) 2001-2007 Christian Diefer

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

PVOID biosVA;
USHORT smTableLength;

PVOID biosVA2;
int biosVA2Mapped=0;

static PUCHAR smEntryPoint;
static PUCHAR smTable;

static void searchSMBIOSEntryPoint( PVOID biosVA, char* pIOBuffer );
static int checkSMBIOSSignature( PUCHAR smEntryPoint, char* pIOBuffer );
static ULONG smbiosGetStructLength( PUCHAR structPtr );
static ULONG smbiosGetReadableName( char* name, PUCHAR structPtr );
static PCHAR getString( PUCHAR structPtr, unsigned int stringnr );
static VOID analyzeStruct( char* pIOBuffer, PUCHAR structPtr );

/////////////////////////////////////////////////////////////////////////////
void getSMBIOS( char* pIOBuffer )
{
  int status;
  
  *pIOBuffer = '\0';

  if( biosVA == NULL )
  {
    sprintf( pIOBuffer, "*** Error: BIOS segment mapping failed ***" );
    return;         // mapping did not work (?)
  }

#ifdef SMBIOSMSG
  sprintf( &pIOBuffer[strlen(pIOBuffer)], ">>> BIOS segment mapped from PA 0x%08x to VA 0x%08x\r\n", 0xe0000, biosVA );
#endif

  if( smEntryPoint == NULL )
    searchSMBIOSEntryPoint( ((PUCHAR)biosVA)+0x10000, pIOBuffer );    // start search at 0xf0000 instead of 0xe0000

  if( smEntryPoint == NULL )
  {
    sprintf( &pIOBuffer[strlen(pIOBuffer)], "*** Error: No SMBIOS found ***" );
    return;
  }

#ifdef SMBIOSMSG
  sprintf( &pIOBuffer[strlen(pIOBuffer)], ">>> SMBIOS entry point found at VA 0x%08x\r\n", smEntryPoint );
#endif

  status = checkSMBIOSSignature( smEntryPoint, pIOBuffer );
  if( status == 1 )
  {
	  sprintf( &pIOBuffer[strlen(pIOBuffer)], "*** Error: Pointer to DMI structures not found ***" );
    return;
  }
  else if( status == 2 )
  {
    sprintf( &pIOBuffer[strlen(pIOBuffer)], "*** Error: SMBIOS table out of BIOS segment ***" );
    return;
  }

  {
    PUCHAR structPtr = smTable;
    int i;

    // iterate through SMBIOS structures
    for( i=0; i<*(PUSHORT)&smEntryPoint[28]; i++ )
    {
      ULONG x = smbiosGetStructLength( structPtr );

#ifdef SMBIOSMSG
      sprintf( &pIOBuffer[strlen(pIOBuffer)], ">>> Block %d (VA 0x%08x, PA 0x%08x): Type: %d Length:%d(%d) Name:",
               i, 
               structPtr, 
               (structPtr-(PUCHAR)biosVA)+0xe0000,
               structPtr[0], 
               x, 
               structPtr[1] );
      smbiosGetReadableName( &pIOBuffer[strlen(pIOBuffer)], structPtr );
      strcat( pIOBuffer, "\r\n" );
#endif

      analyzeStruct( pIOBuffer, structPtr );

      structPtr += x;
    }
  }
}

/////////////////////////////////////////////////////////////////////////////
void getCPUSpeed( char* pIOBuffer )
{
  int status;
  
  strcpy( pIOBuffer, "---" );

  if( biosVA == NULL )
  {
    sprintf( pIOBuffer, "*** Error: BIOS segment mapping failed ***" );
    return;         // mapping did not work (?)
  }

#ifdef SMBIOSMSG
  sprintf( &pIOBuffer[strlen(pIOBuffer)], ">>> BIOS segment mapped from PA 0x%08x to VA 0x%08x\r\n", 0xe0000, biosVA );
#endif

  if( smEntryPoint == NULL )
    searchSMBIOSEntryPoint( ((PUCHAR)biosVA)+0x10000, pIOBuffer );  // start search at 0xf0000 instead of 0xe0000

  if( smEntryPoint == NULL )
  {
    sprintf( &pIOBuffer[strlen(pIOBuffer)], "*** Error: No SMBIOS found ***" );
    return;
  }

  status = checkSMBIOSSignature( smEntryPoint, pIOBuffer );
  if( status == 1 )
  {
	  sprintf( &pIOBuffer[strlen(pIOBuffer)], "*** Error: Pointer to DMI structures not found ***" );
    return;
  }
  else if( status == 2 )
  {
    sprintf( &pIOBuffer[strlen(pIOBuffer)], "*** Error: SMBIOS table out of BIOS segment ***" );
    return;
  }

  {
    PUCHAR structPtr = smTable;
    int i;

    // iterate through SMBIOS structures
    for( i=0; i<*(PUSHORT)&smEntryPoint[28]; i++ )
    {
      ULONG x = smbiosGetStructLength( structPtr );

#ifdef SMBIOSMSG
      sprintf( &pIOBuffer[strlen(pIOBuffer)], ">>> Block %d (VA 0x%08x, PA 0x%08x): Type: %d Length:%d(%d) Name:",
               i, 
               structPtr, 
               (structPtr-(PUCHAR)biosVA)+0xe0000,
               structPtr[0], 
               x, 
               structPtr[1] );
      smbiosGetReadableName( &pIOBuffer[strlen(pIOBuffer)], structPtr );
      strcat( pIOBuffer, "\r\n" );
#endif

      if( structPtr[0] == 0x04 )    // CPU info
      {
        sprintf( pIOBuffer, "%d", *(PUSHORT)&structPtr[22] );
        break;
      }

      structPtr += x;
    }
  }
}

/////////////////////////////////////////////////////////////////////////////
static void searchSMBIOSEntryPoint( PVOID addr, char* pIOBuffer )
{
  PUCHAR temp;
  
  smEntryPoint = NULL;

  // search for SMBIOS entry point in BIOS segment
  for( temp=(PUCHAR)addr; temp<(((PUCHAR)addr)+0x10000); temp+=4 )
	{
    if( temp[0]==0x5f
    &&  temp[1]==0x53
    &&  temp[2]==0x4d
    &&  temp[3]==0x5f )   // SMBIOS_MAGIC_DWORD found
    {
      PUCHAR i;
      UCHAR checksum = 0;
      UCHAR length = ((PUCHAR)temp)[5];

      // calculate checksum for entry point structure
      for( i = (PUCHAR)temp; i < ((PUCHAR)temp) + length; i++ )
        checksum += *i;

      if( checksum == 0 )   // gotcha
      {
        smEntryPoint = (PUCHAR)temp;

#ifdef SMBIOSMSG
        sprintf( &pIOBuffer[strlen(pIOBuffer)], ">>> Detected SMBIOS V%d.%d at VA 0x%08x (PA 0x%08x)\r\n",
                 smEntryPoint[6],
                 smEntryPoint[7],
                 smEntryPoint,
                 (smEntryPoint-(PUCHAR)biosVA)+0xe0000 );
#endif
        break;
      }
    }
  }
}

/////////////////////////////////////////////////////////////////////////////
static int checkSMBIOSSignature( PUCHAR smEntryPoint, char* pIOBuffer )
{
  if( strncmp((PCHAR)&smEntryPoint[16],DMI_STRING,sizeof(DMI_STRING)-1) )
    return 1;

  // retrieve table data
//#ifdef _WIN64
  smTable = (PUCHAR)(ULONG_PTR)*(PULONG)&smEntryPoint[24];
//#else
//  smTable = (PUCHAR)(ULONG_PTR)*(PULONG)&smEntryPoint[24];
//#endif

  smTableLength = *(PUSHORT)&smEntryPoint[22];

#ifdef SMBIOSMSG
  sprintf( &pIOBuffer[strlen(pIOBuffer)], ">>> SMBIOS Table at PA 0x%08x, Length %d (%d structures)\r\n",
           smTable,
           smTableLength,
           *(PUSHORT)&smEntryPoint[28] );
#endif

  // SMBIOS table reaches out of currently mapped BIOS segment => remap BIOS segment
#ifdef _WIN64
  if( smTable-((ULONG64)0xf0000) > ((PUCHAR)(ULONG64)0xffff) )
#else
  if( ((ULONG)(smTable-(ULONG)0xf0000)) > (ULONG)0xffff )
#endif
  {
#ifdef SMBIOSMSG
    sprintf( &pIOBuffer[strlen(pIOBuffer)], ">>> SMBIOS Table out of BIOS area => mapping new area\r\n" );
#endif

    if( !biosVA2Mapped )
    {
#ifdef _WIN64
      biosPA2.LowPart = (ULONG)(((ULONG64)smTable)&0xffffffff);
      biosPA2.HighPart = (ULONG)(((ULONG64)smTable)>>32);
#else
      biosPA2.LowPart = (ULONG)smTable;
      biosPA2.HighPart = 0;
#endif
      biosVA2 = MmMapIoSpace( biosPA2, smTableLength, MmNonCached );
      biosVA2Mapped = -1;

#ifdef SMBIOSMSG
      sprintf( &pIOBuffer[strlen(pIOBuffer)], ">>> SMBIOS area mapped to VA 0x%08x\r\n", biosVA2 );
#endif
    }
    
    smTable = (PUCHAR)biosVA2;
  }
  else
  {
#ifdef _WIN64
    smTable = smTable - (ULONG64)0xe0000;  // table offset within BIOS area
    smTable = smTable + (ULONG64)biosVA;   // translate offset to VA (PA -> VA)
#else
    smTable = smTable - (ULONG)0xe0000;  // table offset within BIOS area
    smTable = smTable + (ULONG)biosVA;   // translate offset to VA (PA -> VA)
#endif
  }

  return 0;  
}

/////////////////////////////////////////////////////////////////////////////
static VOID analyzeStruct( char* pIOBuffer, PUCHAR structPtr )
{
  if( structPtr[0] == 0x00 )    // BIOS info
  {
    sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "BIOS vendor: %s\r\n", getString(structPtr,structPtr[4]) );
    sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "BIOS version: %s\r\n", getString(structPtr,structPtr[5]) );
//    sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "BIOS segment: %x\r\n", *(PUSHORT)&structPtr[6] );
    sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "BIOS date: %s\r\n", getString(structPtr,structPtr[8]) );
//    sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "ROM size: %s\r\n", getString(structPtr,structPtr[9]) );
  }
  else if( structPtr[0] == 0x01 )    // System info
  {
    sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "System manufacturer: %s\r\n", getString(structPtr,structPtr[4]) );
    sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "Product name: %s\r\n", getString(structPtr,structPtr[5]) );
//    sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "Version: %s\r\n", getString(structPtr,structPtr[6]) );
    sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "Service tag: %s\r\n", getString(structPtr,structPtr[7]) );
  }
  else if( structPtr[0] == 0x02 )    // Base board info
  {
/*
    sprintf( (PCHAR)&pIOBuffer[(PCHAR)strlen(pIOBuffer)], "System manufacturer: %s\r\n", getString(structPtr,structPtr[4]) );
    sprintf( (PCHAR)&pIOBuffer[(PCHAR)strlen(pIOBuffer)], "Product name: %s\r\n", getString(structPtr,structPtr[5]) );
    sprintf( (PCHAR)&pIOBuffer[(PCHAR)strlen(pIOBuffer)], "Version: %s\r\n", getString(structPtr,structPtr[6]) );
    sprintf( &(PCHAR)pIOBuffer[(PCHAR)strlen(pIOBuffer)], "Serial number: %s\r\n", getString(structPtr,structPtr[7]) );
*/
  }
  else if( structPtr[0] == 0x03 )    // system enclosure info
  {
    sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "Asset tag: %s\r\n", getString(structPtr,structPtr[8]) );
/*
    sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "Manufacturer: %s\r\n", getString(structPtr,structPtr[4]) );
    sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "Type: %s\r\n", getString(structPtr,structPtr[5]) );
    sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "Version: %s\r\n", getString(structPtr,structPtr[6]) );
    sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "Serial number: %s\r\n", getString(structPtr,structPtr[7]) );
    sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "Bootup state: %s\r\n", getString(structPtr,structPtr[9]) );
    sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "Power supply state: %s\r\n", getString(structPtr,structPtr[10]) );
    sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "Thermal state: %s\r\n", getString(structPtr,structPtr[11]) );
*/
  }
  else if( structPtr[0] == 0x04 )    // CPU info
  {
//    sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "CPU socket type: %s\r\n", getString(structPtr,structPtr[4]) );
    sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "CPU manufacturer: %s\r\n", getString(structPtr,structPtr[7]) );
//    sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "CPU version: %s\r\n", getString(structPtr,structPtr[16]) );
    sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "Current CPU speed: %d MHz\r\n", *(PUSHORT)&structPtr[22] );
    sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "External clock: %d MHz\r\n", *(PUSHORT)&structPtr[18] );
    sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "Max. CPU speed supported by board: %d MHz\r\n", *(PUSHORT)&structPtr[20] );
  }
  else if( structPtr[0] == 0x0a )   // onboard devices
  {
    switch( structPtr[4] & 0x7f )
    {
      case 3: sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "Video card: %s\r\n", getString(structPtr,structPtr[5]) );
              break;

      case 5: sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "Ethernet card: %s\r\n", getString(structPtr,structPtr[5]) );
              break;

      case 7: sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "Sound card: %s\r\n", getString(structPtr,structPtr[5]) );
              break;
    }
  }
  else if( structPtr[0] == 7 )    // cache info
  {
    unsigned type = (*(PUSHORT)&structPtr[5]) & 3;
    unsigned gran = ((*(PUSHORT)&structPtr[9]) & 0x8000) ? 64 : 1;
    unsigned loc = ((*(PUSHORT)&structPtr[5]) & 0x60)>>5;

    sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "L%d cache size: %d KB (", type+1, ((*(PUSHORT)&structPtr[9]) & 0x7fff) * gran );

    switch( loc )
    {
      case 0: strcat( (PCHAR)pIOBuffer, "internal" );
              break;

      case 1: strcat( (PCHAR)pIOBuffer, "external" );
              break;

      case 2: strcat( (PCHAR)pIOBuffer, "reserved" );
              break;

      default:strcat( (PCHAR)pIOBuffer, "unknown" );
    }

    strcat( (PCHAR)pIOBuffer, ")\r\n" );
  }
  else if( structPtr[0] == 17 )    // memory info
  {
    sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "SODIMM Socket '%s': %d MB", getString(structPtr,structPtr[16]), *(PUSHORT)&structPtr[12] );
    if( structPtr[1]>=0x17 && *(PUSHORT)&structPtr[21] )
      sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], " (%d MHz)", *(PUSHORT)&structPtr[21] );

    strcat( (PCHAR)pIOBuffer, "\r\n" );
  }
  else if( structPtr[0] == 22 )    // battery info
  {
    sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "Battery manufacturer: %s\r\n", getString(structPtr,structPtr[5]) );
    sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "Battery device name: %s\r\n",  getString(structPtr,structPtr[8]) );
    sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "Battery location: %s\r\n", getString(structPtr,structPtr[4]) );

    if( structPtr[6] )
      sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "Battery manufacturing date: %s\r\n", getString(structPtr,structPtr[6]) );
    else if( structPtr[1] >= 19 )
    {
      USHORT x = *(PUSHORT)&structPtr[18];

      sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "Battery manufacturing date: %d/%d/%d (SBDS)\r\n",
                (x&0x1e0)>>5, x&0x1f, ((x&0xfe00)>>9)+1980 );
    }

    if( structPtr[7] )
      sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "Battery serial no: %s\r\n", getString(structPtr,structPtr[7]) );
    else if( structPtr[1] >= 17 )
      sprintf( (PCHAR)&pIOBuffer[strlen((PCHAR)pIOBuffer)], "Battery serial no: %d (SBDS)\r\n",  *(PUSHORT)&structPtr[18] );
  }
}

/////////////////////////////////////////////////////////////////////////////
static ULONG smbiosGetStructLength( PUCHAR structPtr )
{
  // jump to string list
  PUCHAR ptr = ((PUCHAR)structPtr) + structPtr[1];
  ULONG len = structPtr[1];

  // search for the end of string list
  while( ptr[0]!=0x00 || ptr[1]!=0x00 )
  {
    ptr++;
    len++;
  }

  ptr += 2;			// terminating 0x0000 should be included
  len += 2;
  
  return len;
}

/////////////////////////////////////////////////////////////////////////////
static ULONG smbiosGetReadableName( char* name, PUCHAR structPtr )
{
  switch( structPtr[0] )
  {
    case 0: return sprintf (name, "%s", RD_BIOS);
    case 1: return sprintf (name, "%s", RD_SYSTEM);
		case 2: return sprintf (name, "%s", RD_BASEBOARD);
		case 3: return sprintf (name, "%s", RD_ENCLOSURE);
		case 4: return sprintf (name, "%s", RD_PROCESSOR);
		case 5: return sprintf (name, "%s", RD_MEMCTRL);
		case 6: return sprintf (name, "%s", RD_MEMMOD);
		case 7: return sprintf (name, "%s", RD_CACHE);
		case 8: return sprintf (name, "%s", RD_PORT);
		case 9: return sprintf (name, "%s", RD_SLOT);
		case 10: return sprintf (name, "%s", RD_ONBOARD);
		case 11: return sprintf (name, "%s", RD_OEMSTRINGS);
		case 12: return sprintf (name, "%s", RD_SYSTEMCONFIG);
		case 13: return sprintf (name, "%s", RD_BIOSLANG);
		case 14: return sprintf (name, "%s", RD_GROUPASSOC);
		case 15: return sprintf (name, "%s", RD_EVENTLOG);
		case 16: return sprintf (name, "%s", RD_MEMARRAY);
		case 17: return sprintf (name, "%s", RD_MEMDEV);
		case 18: return sprintf (name, "%s", RD_32MEMERR);
		case 19: return sprintf (name, "%s", RD_MEMMAPPEDADR);
		case 20: return sprintf (name, "%s", RD_MEMMAPPEDDEV);
		case 21: return sprintf (name, "%s", RD_POINTINGDEV);
		case 22: return sprintf (name, "%s", RD_BATTERY);
		case 23: return sprintf (name, "%s", RD_RESET);
		case 24: return sprintf (name, "%s", RD_SECURITY);
		case 25: return sprintf (name, "%s", RD_PWRCTRL);
		case 26: return sprintf (name, "%s", RD_VOLTAGE);
		case 27: return sprintf (name, "%s", RD_COOLINGDEV);
		case 28: return sprintf (name, "%s", RD_TEMP);
		case 29: return sprintf (name, "%s", RD_CURRENT);
		case 30: return sprintf (name, "%s", RD_RMTACCESS);
		case 31: return sprintf (name, "%s", RD_BIS);
		case 32: return sprintf (name, "%s", RD_BOOT_INFO);
		case 33: return sprintf (name, "%s", RD_64MEMERR);
		case 34: return sprintf (name, "%s", RD_MANAGDEV);
		case 35: return sprintf (name, "%s", RD_MANAGDEVCOMP);
		case 36: return sprintf (name, "%s", RD_MANAGDEVTHRESH);
		case 37: return sprintf (name, "%s", RD_MEMCHANNEL);
		case 38: return sprintf (name, "%s", RD_IPMI);
		case 39: return sprintf (name, "%s", RD_PWRSUP);
		case 126: return sprintf (name, "%s", RD_INACTIVE);
		case 127: return sprintf (name, "%s", RD_EOT);
		default: return sprintf (name, "%d", structPtr[0] );
  }
}

/////////////////////////////////////////////////////////////////////////////
static PCHAR getString( PUCHAR structPtr, unsigned int stringnr )
{
	unsigned int str_stringnr = 0;
	unsigned char *ptr = NULL;


  // if we should get string nr 0, we have nothing to do
	if( stringnr == str_stringnr )
    return "--";

	str_stringnr = 1;

  // jump to string list
  ptr = ((PUCHAR)structPtr) + (ULONG)(structPtr[1]);

  // double 0x00 indicates we reached the end of the string list
  if( ptr[0]==0x00 && ptr[1]==0x00 )
    return "--";

  while( str_stringnr < stringnr )		// as long as we have not found our string
	{
    ptr = ptr + strlen((PCHAR)ptr) + 1;
    if( ptr[0] == 0x00 )
      return NULL;	// this means we have a double 0x00 and have reached the end of the string list

    str_stringnr++;
	}

  return (PCHAR)ptr;
}
