/*
   I8kfanGUI V3.0
   Copyright(c) 2001-2006 Christian Diefer
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2 as 
   published by the Free Software Foundation.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "stdafx.h"
#include <stdio.h>
#include <math.h>
#include <limits.h>
#include "cpuspeed.h"
#include "cpuid.h"

static int firstrun = 1;
static ushort sproc;
static DWORD sfeat;


//---------------------------------------------------
// CPU-Speed mit TSC und Performance-Counter ermitteln (Variante 1)
//---------------------------------------------------
int cpuSpeedHT()
{
	UINT64 t0;      // ticks at beginning of test
  UINT64 t1;      // ticks at end of test
  UINT64 stamp0;  // cycles at beginning of test
  UINT64 stamp1;  // cycles at end of test
  UINT64 countFreq;
  UINT64 elapsedCycles;
  UINT64 elapsedTicks;
  double elapsedTime;
	HANDLE hProcess = GetCurrentProcess();
	HANDLE hThread = GetCurrentThread();

  if( firstrun )    // init once (at first call only)
  {
    sproc = wincpuid();
    sfeat = wincpufeatures();
    firstrun = 0;
  }

  if( !(sfeat & TSC_SUPPORT) )    // no TSC support
		return 0;

	// get the frequency of the high-resolution performance counter
	if( !QueryPerformanceFrequency((LARGE_INTEGER*)&countFreq) )
		return 0;

	// get current process & thread priorities for later restoration
  ULONG dwCurPriorityClass = GetPriorityClass( hProcess );
	int iCurThreadPriority = GetThreadPriority( hThread );

	// set this process & thread to max priorities so we don't get interrupted too often
  if( dwCurPriorityClass != 0 )
    SetPriorityClass( hProcess, REALTIME_PRIORITY_CLASS );

  if( iCurThreadPriority != THREAD_PRIORITY_ERROR_RETURN )
    SetThreadPriority( hThread, THREAD_PRIORITY_TIME_CRITICAL );

  // get starting ticks
  QueryPerformanceCounter( (LARGE_INTEGER*)&t0 );

	// serialize for in-order-execution and get starting cycles 
#ifdef _WIN64
	readRDTSC( &stamp0 );
#else
	__asm 
  { 
    push ebx
    xor eax,eax
    cpuid
    rdtsc
    mov dword ptr [stamp0],eax
    mov dword ptr [stamp0+4],edx
    pop ebx
	}
#endif

	Sleep( 50 );    // wait for 50msec to let the ticks and cycles counters advance

  // get ending ticks
  QueryPerformanceCounter( (LARGE_INTEGER*)&t1 );

	// serialize again for in-order-execution and get ending cycles 
#ifdef _WIN64
  readRDTSC( &stamp1 );
#else
  __asm 
	{
      push ebx
      xor eax,eax
      cpuid
      rdtsc
      mov dword ptr [stamp1],eax
      mov dword ptr [stamp1+4],edx
      pop ebx
	}
#endif

  // restore process & thread priorities
	SetThreadPriority( hThread, iCurThreadPriority );
	SetPriorityClass( hProcess, dwCurPriorityClass );

  // now compute the frequency...
	elapsedCycles = stamp1 - stamp0;
	elapsedTicks = t1 - t0;
  elapsedTime = (double)elapsedTicks / (double)countFreq;

  // MHz = elapsed cycles / elapsed time / 1 million
  return (int)((UINT64)(((double)elapsedCycles) / elapsedTime) / 1000000);
}

//-----------------------------------------------------
// CPU-Speed mit TSC und Performance-Counter ermitteln (Variante 2)
//
// Rckgabewerte: 0=alles ok
//                1=kann PerformanceFrequency nicht lesen
//                2=(ticks==0) => nicht ausfhrbar auf dieser CPU
//                3=TSC nicht vorhanden auf dieser CPU
//-----------------------------------------------------
int getRDTSCSpeed( struct FREQ_INFO* info )
{
  UINT64 stamp0=0, stamp1=0;        // Time Stamp Variable for beginning and end of test
  UINT64 t0, t1;                    // Variables for High-Resolution Performance Counter reads
  UINT64 freq=0;                    // Most current frequ. calculation
  UINT64 freq2=0;                   // 2nd most current frequ. calc.
  UINT64 freq3=0;                   // 3rd most current frequ. calc.
  UINT64 total;                     // Sum of previous three frequency calculations
  int tries=0;                      // Number of times a calculation has been made on this call to cpuspeed
  UINT64 total_cycles=0, cycles=0;  // Clock cycles elapsed during test
  UINT64 total_ticks=0, ticks;      // Microseconds elapsed during test
  UINT64 count_freq;                // High Resolution Performance Counter frequency

  if( firstrun )    // init once (at first call only)
  {
    sproc = wincpuid();
    sfeat = wincpufeatures();
    firstrun = 0;
  }

  memset( info, 0x00, sizeof(*info) );

  if( !(sfeat & TSC_SUPPORT) )    // no TSC support
		return 3;

  int iPriority;
  HANDLE hThread = GetCurrentThread();

  if( !QueryPerformanceFrequency((LARGE_INTEGER*)&count_freq) )
    return 1;

  // On processors supporting the Read Time Stamp opcode, compare elapsed time on the 
  //   High-Resolution Counter with elapsed cycles on the Time Stamp Register.
    
  // This do loop runs up to 20 times or until the average of the previous three calculated frequencies is 
  //   within 1 MHz of each of the individual calculated frequencies. This resampling increases the 
  //   accuracy of the results since outside factors could affect this calculation
  do 
  {          
    tries++;        // Increment number of times sampled on this call to cpuspeed
            
    freq3 = freq2;  // Shift frequencies back to make room for new frequency measurement
    freq2 = freq;

    QueryPerformanceCounter( (LARGE_INTEGER*)&t0 );   // Get high-resolution performance counter time

    t1 = t0;

    iPriority = GetThreadPriority( hThread );
    if( iPriority != THREAD_PRIORITY_ERROR_RETURN )
      SetThreadPriority( hThread, THREAD_PRIORITY_TIME_CRITICAL );

    while( (t1-t0) < 50 ) 
    {
      // Loop until 50 ticks have passed since last read of hi-res counter. 
      //   This accounts for overhead later.
      QueryPerformanceCounter( (LARGE_INTEGER*)&t1 );

#ifdef _WIN64
      readRDTSC( &stamp0 );
#else
      _asm {
        push eax
        push ebx
        push edx
        xor eax,eax
        cpuid
        rdtsc
		    mov dword ptr [stamp0],eax
		    mov dword ptr [stamp0+4],edx
		    pop edx
        pop ebx
        pop eax
      }
#endif
    }   

    t0 = t1;  // reset initial time

    while( (t1-t0) < 1000 ) 
    {
      // Loop until 1000 ticks have passed since last read of hi-res counter. 
      //   This allows for elapsed time for sampling.
      QueryPerformanceCounter( (LARGE_INTEGER*)&t1 );

#ifdef _WIN64
      readRDTSC( &stamp1 );
#else
      _asm {
        push eax
        push ebx
        push edx
        xor eax,eax
        cpuid
        rdtsc
		    mov dword ptr [stamp1],eax
		    mov dword ptr [stamp1+4],edx
		    pop edx
        pop ebx
        pop eax
      }
#endif
    }

    // Reset priority
    if( iPriority != THREAD_PRIORITY_ERROR_RETURN )
      SetThreadPriority( hThread, iPriority );

    // Number of internal clock cycles is difference between two time stamp readings.
    cycles = stamp1 - stamp0;   

    // Number of external ticks is difference between two hi-res counter reads.
    ticks = t1 - t0;


    // Note that some seemingly arbitrary mulitplies and divides are done below. This is to maintain a 
    //   high level of precision without truncating the most significant data. According to what value 
    //   ITERATIIONS is set to, these multiplies and divides might need to be shifted for optimal precision.

    // Convert ticks to hundred thousandths of a tick
    ticks = ticks * 100000;

    // Hundred Thousandths of a Ticks / ( 10 ticks/second ) = microseconds (us)
    ticks = ticks / (count_freq/10);      

    total_ticks += ticks;
    total_cycles += cycles;

    if( ticks%count_freq > count_freq/2 )
      ticks++;            // Round up if necessary

    if( ticks == 0 )
      return 2;

    freq = cycles/ticks;    // Cycles / us  = MHz

    if( cycles%ticks > ticks/2 )
      freq++;             // Round up if necessary

    total = ( freq + freq2 + freq3 );     // Total last three frequency calculations
  } while( (tries < 3 ) ||        
           (tries < 20)&&
           ((abs(3 * (long)freq - (long)total) > 3*TOLERANCE )||
            (abs(3 * (long)freq2 - (long)total) > 3*TOLERANCE )||
            (abs(3 * (long)freq3 - (long)total) > 3*TOLERANCE )));
    // ... Compare last three calculations to average of last three calculations.                           

  // Try one more significant digit.
  freq3 = ( total_cycles * 10 ) / total_ticks;
  freq2 = ( total_cycles * 100 ) / total_ticks;

  if( freq2 - (freq3 * 10) >= ROUND_THRESHOLD )
    freq3++;

  info->raw_freq = total_cycles / total_ticks;
  info->norm_freq = info->raw_freq;

  freq = info->raw_freq * 10;
  if( (freq3 - freq) >= ROUND_THRESHOLD )
    info->norm_freq++;

  return 0;
}
